# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2016-12-22 18:21
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('onfido', '0010_convert_raw_fields_to_jsonb'),
    ]

    operations = [
        migrations.AlterField(
            model_name='check',
            name='status',
            field=models.CharField(blank=True, choices=[('Check', (('in_progress', 'In progress'), ('awaiting_applicant', 'Awaiting applicant'), ('complete', 'Complete'), ('withdrawn', 'Withdrawn'), ('paused', 'Paused'), ('reopened', 'Reopened'))), ('Report', (('awaiting_data', 'Awaiting data'), ('awaiting_approval', 'Awaiting approval'), ('complete', 'Complete'), ('withdrawn', 'Withdrawn'), ('paused', 'Paused'), ('cancelled', 'Cancelled')))], db_index=True, help_text='The current state of the check / report (from API).', max_length=20, null=True),
        ),
        migrations.AlterField(
            model_name='report',
            name='status',
            field=models.CharField(blank=True, choices=[('Check', (('in_progress', 'In progress'), ('awaiting_applicant', 'Awaiting applicant'), ('complete', 'Complete'), ('withdrawn', 'Withdrawn'), ('paused', 'Paused'), ('reopened', 'Reopened'))), ('Report', (('awaiting_data', 'Awaiting data'), ('awaiting_approval', 'Awaiting approval'), ('complete', 'Complete'), ('withdrawn', 'Withdrawn'), ('paused', 'Paused'), ('cancelled', 'Cancelled')))], db_index=True, help_text='The current state of the check / report (from API).', max_length=20, null=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from decimal import Decimal

from django.contrib.auth import get_user_model
from django.test import TestCase
from django.utils import timezone

from ..admin import (
    Applicant,
    Check,
    Event,
    EventsMixin,
    RawMixin,
    ResultMixin,
    UserMixin,
)
from ..compat import mock


class ResultMixinTests(TestCase):

    """onfido.admin.ResultMixin tests."""

    @mock.patch.object(Check, 'mark_as_clear')
    def test__events(self, mock_clear):

        def request():
            request = mock.Mock()
            request.user = get_user_model()()
            return request

        check = Check()
        request = request()
        mixin = ResultMixin()
        mixin.mark_as_clear(request, [check])
        mock_clear.assert_called_once_with(request.user)


class EventsMixinTests(TestCase):

    """onfido.admin.EventsMixin tests."""

    @mock.patch.object(Check, 'events')
    def test__events(self, mock_events):
        now = timezone.now()
        event = Event(
            onfido_id='foo',
            action='test.action',
            resource_type='bar',
            status='in_progress',
            completed_at=now.isoformat(),
            received_at=now.isoformat(),
        ).save()
        mock_events.return_value = [event]
        mixin = EventsMixin()
        check = Check()
        html = mixin._events(check)
        self.assertEqual(html, '<ul><li>{}: test.action</li></ul>'.format(now.date().isoformat()))


class RawMixinTests(TestCase):

    """onfido.admin.RawMixin tests."""

    def test__raw(self):
        mixin = RawMixin()
        obj = Applicant(raw={'foo': 'bar'})
        html = mixin._raw(obj)
        self.assertEqual(html, '<code>{<br>&nbsp;&nbsp;&nbsp;&nbsp;"foo":&nbsp;"bar"<br>}</code>')

        # test with Decimal (stdlib json won't work) and unicode
        obj = Applicant(raw={'foo': Decimal(1.0), 'bar': 'åß∂ƒ©˙∆'})
        html = mixin._raw(obj)


class UserMixinTests(TestCase):

    """onfido.admin.UserMixin tests."""

    def test__user(self):

        def assertUser(first_name, last_name, expected):
            mixin = UserMixin()
            user = get_user_model()(first_name=first_name, last_name=last_name)
            obj = mock.Mock(user=user)
            self.assertEqual(mixin._user(obj), expected)

        assertUser('fred', 'flintstone', 'Fred Flintstone')
        assertUser('', '', '')
        assertUser('fredå', 'flintstone', 'Fredå Flintstone')
        # this fails, something that title() does to unicode?
        with self.assertRaises(AssertionError):
            self.assertEqual('ƒ', 'ƒ'.title())
        assertUser('ƒredå', '', 'ƒredå'.title())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('onfido', '0008_rename_event_created_at'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='event',
            options={'ordering': ['completed_at']},
        ),
        migrations.AddField(
            model_name='check',
            name='is_clear',
            field=models.NullBooleanField(default=None, help_text="True if the check / report is 'clear' (via API or manual override)."),
        ),
        migrations.AddField(
            model_name='report',
            name='is_clear',
            field=models.NullBooleanField(default=None, help_text="True if the check / report is 'clear' (via API or manual override)."),
        ),
    ]



# -*- coding: utf-8 -*-
from django.apps import AppConfig


class OnfidoAppConfig(AppConfig):

    """AppConfig for Django-Onfido."""

    name = 'onfido'
    verbose_name = "Onfido"
    configs = []

    def ready(self):
        """Validate config and connect signals."""
        super(OnfidoAppConfig, self).ready()



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-11-15 22:50
from __future__ import unicode_literals

from django.db import migrations

from ..db.fields import JSONField


class Migration(migrations.Migration):

    dependencies = [
        ('onfido', '0009_add_is_clear_field'),
    ]

    operations = [
        migrations.AlterField(
            model_name='applicant',
            name='raw',
            field=JSONField(blank=True, help_text='The raw JSON returned from the API.', null=True),
        ),
        migrations.AlterField(
            model_name='check',
            name='raw',
            field=JSONField(blank=True, help_text='The raw JSON returned from the API.', null=True),
        ),
        migrations.AlterField(
            model_name='report',
            name='raw',
            field=JSONField(blank=True, help_text='The raw JSON returned from the API.', null=True),
        ),
        migrations.AlterField(
            model_name='event',
            name='raw',
            field=JSONField(blank=True, help_text='The raw JSON returned from the API.', null=True),
        ),
    ]

# Below is an example of a full parallel conversion that creates a new field
# copies across the data from the old field, drops the old field and renames
# the new field to the original name. Only required if for some reason a
# straight ALTER statement won't work on existing data. For reference, the
# SQL is: ALTER TABLE "onfido_report" ALTER COLUMN "raw" TYPE jsonb USING "raw"::jsonb;


# def migrate_old_to_new(klass):
#     """Copy the JSON in the TextField across to the new JSONField."""
#     logger.info("Converting {}".format(klass))
#     for obj in klass.objects.order_by('-id'):
#         obj.new_raw = obj.raw
#         logger.info("Saving {}.{}".format(klass, obj.id))
#         obj.save(update_fields=['new_raw'])


# def migrate_applicants(apps, schema_editor):
#     migrate_old_to_new(apps.get_model("onfido", "Applicant"))


# def migrate_checks(apps, schema_editor):
#     migrate_old_to_new(apps.get_model("onfido", "Check"))


# def migrate_reports(apps, schema_editor):
#     migrate_old_to_new(apps.get_model("onfido", "Report"))


# def migrate_events(apps, schema_editor):
#     migrate_old_to_new(apps.get_model("onfido", "Event"))


# class Migration(migrations.Migration):

#     dependencies = [
#         ('onfido', '0009_add_is_clear_field'),
#     ]

#     operations = [
#         # create the new fields
#         migrations.AddField(
#             model_name='applicant',
#             name='new_raw',
#             field=JSONField(blank=True, help_text='The raw JSON returned from the API.', null=True),
#         ),
#         migrations.AddField(
#             model_name='check',
#             name='new_raw',
#             field=JSONField(blank=True, help_text='The raw JSON returned from the API.', null=True),
#         ),
#         migrations.AddField(
#             model_name='report',
#             name='new_raw',
#             field=JSONField(blank=True, help_text='The raw JSON returned from the API.', null=True),
#         ),
#         migrations.AddField(
#             model_name='event',
#             name='new_raw',
#             field=JSONField(blank=True, help_text='The raw JSON returned from the API.', null=True),
#         ),
#         # 2. migrate the data across from the old fields
#         migrations.RunPython(migrate_applicants),
#         migrations.RunPython(migrate_checks),
#         migrations.RunPython(migrate_reports),
#         migrations.RunPython(migrate_events),
#         # 3. remove the old fields
#         migrations.RemoveField(model_name='applicant', name='raw'),
#         migrations.RemoveField(model_name='check', name='raw'),
#         migrations.RemoveField(model_name='report', name='raw'),
#         migrations.RemoveField(model_name='event', name='raw'),
#         # 4. rename the new field to the same name as the old field
#         migrations.RenameField(model_name='applicant', old_name='new_raw', new_name='raw'),
#         migrations.RenameField(model_name='check', old_name='new_raw', new_name='raw'),
#         migrations.RenameField(model_name='report', old_name='new_raw', new_name='raw'),
#         migrations.RenameField(model_name='event', old_name='new_raw', new_name='raw'),
#     ]




# -*- coding: utf-8 -*-
from django.test import TestCase

from ..api import (
    get,
    post,
    _url,
    _headers,
    _respond,
    ApiError,
)
from ..compat import mock


class ApiTests(TestCase):

    """onfido.api module tests."""

    def test__headers(self):
        """Test the _headers function return valid dict."""
        self.assertEqual(
            _headers(api_key="foo"),
            {
                "Authorization": "Token token=foo",
                "Content-Type": "application/json"
            }
        )

    def test__respond(self):
        """Test the _respond function handles 2xx."""
        response = mock.Mock()
        response.status_code = 200
        self.assertEqual(
            _respond(response),
            response.json.return_value
        )
        # non-2xx should raise error
        response.status_code = 400
        response.json.return_value = {
            "error": {
                "message": "foo",
                "type": "bar"
            }
        }
        self.assertRaises(ApiError, _respond, response)

    @mock.patch('requests.get')
    @mock.patch('onfido.api._headers')
    def test_get(self, mock_headers, mock_get):
        """Test the get function calls API."""
        response = mock.Mock()
        response.status_code = 200
        headers = mock_headers.return_value
        mock_get.return_value = response
        self.assertEqual(get('/'), response.json.return_value)
        mock_get.assert_called_once_with(_url('/'), headers=headers)

    @mock.patch('requests.post')
    @mock.patch('onfido.api._headers')
    def test_post(self, mock_headers, mock_post):
        """Test the get function calls API."""
        response = mock.Mock()
        response.status_code = 200
        headers = mock_headers.return_value
        data = {"foo": "bar"}
        mock_post.return_value = response
        self.assertEqual(post('/', data), response.json.return_value)
        mock_post.assert_called_once_with(_url('/'), headers=headers, json=data)



# -*- coding: utf-8 -*-
import json

from django.contrib.auth import get_user_model
from django.test import TestCase, RequestFactory

from ..compat import mock
from ..models import (
    Applicant,
    Check,
    Report,
    Event,
)
from ..views import status_update


class ViewTests(TestCase):

    """onfido.views module tests."""

    @mock.patch('onfido.decorators.WEBHOOK_TOKEN')
    def test_status_update(self, *args):
        """Test the status_update view function."""
        data = {
            "payload": {
                "resource_type": "check",
                "action": "check.form_completed",
                "object": {
                    "id": "5345badd-f4bf-4240-9f3b-ffb998bda09e",
                    "status": "in_progress",
                    "completed_at": "2016-10-15 11:34:09 UTC",
                    "href": "https://api.onfido.com/v1/applicants/4d390bbd-63c7-4960-8304-a7a04a8051e8/checks/5345badd-f4bf-4240-9f3b-ffb998bda09e"  # noqa
                }
            }
        }
        factory = RequestFactory()

        @mock.patch('onfido.decorators._match', lambda x, y: True)
        def assert_update(data, message):
            request = factory.post(
                '/',
                data=json.dumps(data),
                content_type='application/json'
            )
            response = status_update(request)
            self.assertEqual(response.status_code, 200)
            self.assertEqual(response.content.decode('utf-8'), message)

        # invalid payload
        assert_update({}, 'Unexpected event content.')

        # invalid resource type
        data['payload']['resource_type'] = "foo"
        assert_update(data, 'Unknown resource type.')

        # unknown Check
        data['payload']['resource_type'] = "check"
        with mock.patch.object(Check, 'objects') as mock_manager:
            mock_manager.get.side_effect = Check.DoesNotExist()
            assert_update(data, 'Check not found.')

        # unknown Report
        data['payload']['resource_type'] = "report"
        with mock.patch.object(Report, 'objects') as mock_manager:
            mock_manager.get.side_effect = Report.DoesNotExist()
            assert_update(data, 'Report not found.')

        # unknown exception
        with mock.patch.object(Event, 'parse') as mock_parse:
            mock_parse.side_effect = Exception("foobar")
            assert_update(data, 'Unknown error.')

        # valid payload / object
        mock_check = mock.Mock(spec=Check)
        with mock.patch(
            'onfido.models.Event.resource',
            new_callable=mock.PropertyMock(return_value=mock_check)
        ):
            # mock_resource.return_value = mock_check
            assert_update(data, 'Update processed.')
            mock_check.update_status.assert_called_once()

        # now check that a good payload passes.
        data['payload']['resource_type'] = 'check'
        user = get_user_model().objects.create_user('fred')
        applicant = Applicant(user=user, onfido_id='foo').save()
        check = Check(user=user, applicant=applicant, check_type='standard')
        check.onfido_id = data['payload']['object']['id']
        check.save()

        # validate that the LOG_EVENTS setting is honoured
        with mock.patch.object(Event, 'save') as mock_save:

            with mock.patch('onfido.views.LOG_EVENTS', False):
                assert_update(data, 'Update processed.')
                mock_save.assert_not_called()

            # force creation of event
            with mock.patch('onfido.views.LOG_EVENTS', True):
                assert_update(data, 'Update processed.')
                mock_save.assert_called_once_with()



# -*- coding: utf-8 -*-
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpResponse, HttpResponseForbidden
from django.test import TestCase, RequestFactory

from ..decorators import (
    _hmac,
    _match,
    verify_signature
)
from ..compat import mock

# taken from a real requestbin webhook callback
TEST_WEBHOOK_TOKEN = b'bLFiN4S09FV1nH5G7ZJc3nCYqeMZrHcU'
TEST_REQUEST_SIGNATURE = '32f77520f7b025f15ef3ab55be178667c92827e3'
# HttpRequest.body is a bytestring
TEST_REQUEST_BODY = (
    b'{"payload":{"resource_type":"check","action":"check.form_opened",'
    b'"object":{"id":"923d2717-9abd-4c0b-bc4f-33d769547df5",'
    b'"status":"awaiting_applicant","completed_at":"2016-10-26 16:18:24 UTC",'
    b'"href":"https://api.onfido.com/v1/applicants/4d2ff0dc-3352-4eed-9b7c-'
    b'86ac7b619e4e/checks/923d2717-9abd-4c0b-bc4f-33d769547df5"}}}'
)


class DecoratorTests(TestCase):

    """onfido.decorator module tests."""

    def setUp(self):
        self.factory = RequestFactory()

    def get_request(self, body=TEST_REQUEST_BODY, signature=TEST_REQUEST_SIGNATURE):
        """Create test request."""
        if signature:
            return self.factory.post(
                '/', body,
                content_type='application/json',
                HTTP_X_SIGNATURE=signature
            )
        else:
            return self.factory.post(
                '/', body,
                content_type='application/json'
            )

    def test__hmac(self):
        """Test the _hmac function."""
        self.assertEqual(
            _hmac(TEST_WEBHOOK_TOKEN, TEST_REQUEST_BODY),
            TEST_REQUEST_SIGNATURE
        )

    def test__match(self):
        """Test the _match function."""

        def match(token, body):
            request = self.get_request(body)
            return _match(token, request)

        self.assertTrue(match(TEST_WEBHOOK_TOKEN, TEST_REQUEST_BODY))
        self.assertFalse(match('foo', TEST_REQUEST_BODY))
        self.assertFalse(match(TEST_WEBHOOK_TOKEN, 'bar'))

        # test with a good request, but various HMAC error
        with mock.patch('onfido.decorators._hmac') as mock_hmac:
            mock_hmac.side_effect = KeyError()
            self.assertFalse(match(TEST_WEBHOOK_TOKEN, TEST_REQUEST_BODY))
            mock_hmac.side_effect = Exception("We are f-secure")
            self.assertFalse(match(TEST_WEBHOOK_TOKEN, TEST_REQUEST_BODY))

        # test bad request (missing the X-Signature header)
        request = self.get_request(signature=None)
        self.assertFalse(_match(TEST_WEBHOOK_TOKEN, request))

    @mock.patch('onfido.decorators._match')
    def test_verify_signature(self, mock_match):
        """Test the view function decorator itself."""

        @verify_signature()
        def request_function(request):
            """Fake view function that just returns a 200."""
            return HttpResponse()

        request = self.get_request()

        # import module as we want to manipulate the WEBHOOK_TOKEN
        from .. import decorators

        # in TEST_MODE we return immediately
        decorators.TEST_MODE = True
        decorators.WEBHOOK_TOKEN = None
        mock_match.reset_mock()
        request_function(request)
        mock_match.assert_not_called()

        # without a WEBHOOK_TOKEN we should get an error
        decorators.TEST_MODE = False
        decorators.WEBHOOK_TOKEN = None
        self.assertRaises(ImproperlyConfigured, request_function, request)
        mock_match.assert_not_called()

        # mock out _hmac so we can force a pass / fail
        decorators.TEST_MODE = False
        decorators.WEBHOOK_TOKEN = TEST_WEBHOOK_TOKEN
        mock_match.return_value = True
        self.assertIsInstance(request_function(request), HttpResponse)
        mock_match.return_value = False
        self.assertIsInstance(request_function(request), HttpResponseForbidden)



# -*- coding: utf-8 -*-
"""
Basic wire operations with the API - GET/POST/PUT.

This is a simple wrapper around requests.

"""
import logging

import requests

from .compat import urlparse
from .settings import (
    API_ROOT,
    API_KEY
)

logger = logging.getLogger(__name__)


class ApiError(Exception):

    """Error raised when interacting with the API."""

    def __init__(self, response):
        """Initialise error from response object."""
        data = response.json()
        logger.debug("Onfido API error: {}".format(data))
        super(ApiError, self).__init__(data['error']['message'])
        self.error_type = data['error']['type']


def _url(path):
    """Format absolute API URL."""
    return urlparse.urljoin(API_ROOT, path)


def _headers(api_key=API_KEY):
    """Format request headers."""
    return {
        "Authorization": "Token token={0}".format(api_key),
        "Content-Type": "application/json"
    }


def _respond(response):
    """Process common response object."""
    if not str(response.status_code).startswith('2'):
        raise ApiError(response)
    data = response.json()
    logger.debug("Onfido API response: %s", data)
    return data


def get(href):
    """Make a GET request and return the response as JSON."""
    logger.debug("Onfido API GET request: %s", href)
    return _respond(requests.get(_url(href), headers=_headers()))


def post(href, data):
    """Make a POST request and return the response as JSON."""
    logger.debug("Onfido API POST request: %s: %s", href, data)
    return _respond(requests.post(_url(href), headers=_headers(), json=data))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Applicant',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('onfido_id', models.CharField(help_text='The id returned from the Onfido API.', unique=True, max_length=40)),
                ('created_at', models.DateTimeField(help_text='The timestamp returned from the Onfido API.', null=True, blank=True)),
                ('raw', models.TextField(default='{}', help_text='The raw JSON returned from the API.', null=True, blank=True)),
                ('user', models.OneToOneField(related_name='onfido_applicant', to=settings.AUTH_USER_MODEL, help_text='Django user that maps to this applicant.')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Check',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('onfido_id', models.CharField(help_text='The id returned from the Onfido API.', unique=True, max_length=40)),
                ('created_at', models.DateTimeField(help_text='The timestamp returned from the Onfido API.', null=True, blank=True)),
                ('raw', models.TextField(default='{}', help_text='The raw JSON returned from the API.', null=True, blank=True)),
                ('status', models.CharField(blank=True, max_length=20, null=True, help_text='The current state of the check / report (from API).', choices=[('Check', (('in_progress', 'In progress'), ('awaiting_applicant', 'Awaiting applicant'), ('complete', 'Complete'), ('withdrawn', 'Withdrawn'), ('paused', 'Paused'), ('reopened', 'Reopened'))), ('Report', (('awaiting_data', 'Awaiting data'), ('awaiting_approval', 'Awaiting approval'), ('complete', 'Complete'), ('withdrawn', 'Withdrawn'), ('paused', 'Paused'), ('cancelled', 'Cancelled')))])),
                ('result', models.CharField(blank=True, max_length=20, null=True, help_text='The final result of the check / reports (from API).', choices=[('Check', (('clear', 'Clear'), ('consider', 'Consider'))), ('Report', (('clear', 'Clear'), ('consider', 'Consider'), ('unidentified', 'Unidentified')))])),
                ('updated_at', models.DateTimeField(help_text='The timestamp of the most recent status change (from API).', null=True, blank=True)),
                ('check_type', models.CharField(help_text='See https://documentation.onfido.com/#check-types', max_length=10, choices=[('express', 'Express check'), ('standard', 'Standard check')])),
                ('applicant', models.ForeignKey(related_name='checks', to='onfido.Applicant', help_text='The applicant for whom the check is being made.')),
                ('user', models.ForeignKey(related_name='onfido_checks', to=settings.AUTH_USER_MODEL, help_text='The Django user (denormalised from Applicant to make navigation easier).')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Report',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('onfido_id', models.CharField(help_text='The id returned from the Onfido API.', unique=True, max_length=40)),
                ('created_at', models.DateTimeField(help_text='The timestamp returned from the Onfido API.', null=True, blank=True)),
                ('raw', models.TextField(default='{}', help_text='The raw JSON returned from the API.', null=True, blank=True)),
                ('status', models.CharField(blank=True, max_length=20, null=True, help_text='The current state of the check / report (from API).', choices=[('Check', (('in_progress', 'In progress'), ('awaiting_applicant', 'Awaiting applicant'), ('complete', 'Complete'), ('withdrawn', 'Withdrawn'), ('paused', 'Paused'), ('reopened', 'Reopened'))), ('Report', (('awaiting_data', 'Awaiting data'), ('awaiting_approval', 'Awaiting approval'), ('complete', 'Complete'), ('withdrawn', 'Withdrawn'), ('paused', 'Paused'), ('cancelled', 'Cancelled')))])),
                ('result', models.CharField(blank=True, max_length=20, null=True, help_text='The final result of the check / reports (from API).', choices=[('Check', (('clear', 'Clear'), ('consider', 'Consider'))), ('Report', (('clear', 'Clear'), ('consider', 'Consider'), ('unidentified', 'Unidentified')))])),
                ('updated_at', models.DateTimeField(help_text='The timestamp of the most recent status change (from API).', null=True, blank=True)),
                ('report_type', models.CharField(help_text='The name of the report - see https://documentation.onfido.com/#reports', max_length=20, choices=[('identity', 'Identity report'), ('document', 'Document report'), ('street_level', 'Street level report'), ('facial_similarity', 'Facial similarity report'), ('credit', 'Credit report'), ('criminal_history', 'Criminal history'), ('right_to_work', 'Right to work'), ('ssn_trace', 'SSN trace')])),
                ('onfido_check', models.ForeignKey(related_name='reports', to='onfido.Check', help_text='Check to which this report is attached.')),
                ('user', models.ForeignKey(related_name='onfido_reports', to=settings.AUTH_USER_MODEL, help_text='The Django user (denormalised from Applicant to make navigation easier).')),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



# -*- coding: utf-8 -*-
from copy import deepcopy

from dateutil.parser import parse as date_parse

from django.contrib.auth import get_user_model
from django.test import TestCase

from ..helpers import (
    create_applicant,
    create_check,
    # import from helpers to deter possible dependency issues
    Applicant,
    Check,
    Report
)
from ..compat import mock


CREATE_APPLICANT_RETURN = {
    "id": "a9acefdf-3dc5-4973-aa78-20bd36825b50",
    "created_at": "2016-10-18T16:02:04Z",
    "title": "Mr",
    "first_name": "Fred",
    "last_name": "Flintstone",
    "email": "fred@flintstone.com",
    "country": "GBR",
}


class HelperTests(TestCase):

    """onfido.helper module tests."""

    @mock.patch('onfido.helpers.post')
    def test_create_applicant(self, mock_post):
        """Test the create_applicant function."""
        data = deepcopy(CREATE_APPLICANT_RETURN)
        mock_post.return_value = data
        user = get_user_model().objects.create_user(
            username='fred',
            first_name='Fred',
            last_name='Flintstone',
            email='fred@flintstone.com'
        )
        applicant = create_applicant(user)
        mock_post.assert_called_once_with(
            'applicants',
            {
                'first_name': 'Fred',
                'last_name': 'Flintstone',
                'email': 'fred@flintstone.com',
            }
        )
        self.assertEqual(applicant.onfido_id, data['id'])
        self.assertEqual(applicant.user, user)
        self.assertEqual(applicant.created_at, date_parse(data['created_at']))

    @mock.patch('onfido.helpers.post')
    def test_create_applicant_with_custom_data(self, mock_post):
        """Test the create_applicant function with extra custom POST data."""
        data = deepcopy(CREATE_APPLICANT_RETURN)
        mock_post.return_value = data
        user = get_user_model().objects.create_user(
            username='fred',
            first_name='Fred',
            last_name='Flintstone',
            email='fred@flintstone.com'
        )
        create_applicant(
            user,
            country='GBR',
            dob='2016-01-01',
            gender='male'
        )
        mock_post.assert_called_once_with(
            'applicants',
            {
                'first_name': 'Fred',
                'last_name': 'Flintstone',
                'email': 'fred@flintstone.com',
                'gender': 'male',
                'dob': '2016-01-01',
                'country': 'GBR',
            }
        )

    @mock.patch('onfido.helpers.post')
    def test_create_check(self, mock_post):
        """Test the create_check function."""
        applicant_data = deepcopy(CREATE_APPLICANT_RETURN)
        check_data = {
            "id": "b2b75f66-fffd-45a4-ba15-b1e77a672a9a",
            "created_at": "2016-10-18T16:02:08Z",
            "status": "awaiting_applicant",
            "redirect_uri": None,
            "type": "standard",
            "result": None,
            # "form_uri": "https://onfido.com/information/b2b75f66-fffd-45a4-ba15-b1e77a672a9a",
            "reports": [
                {
                    "created_at": "2016-10-18T16:02:08Z",
                    "id": "08345559-852c-4f47-bf65-f9852bf59c4b",
                    "name": "document",
                    "result": None,
                    "status": "awaiting_data",
                    "variant": "standard",
                },
                {
                    "created_at": "2016-10-18T16:02:08Z",
                    "id": "1ffd3e8a-da71-4674-a245-8b52f1492191",
                    "name": "identity",
                    "result": None,
                    "status": "awaiting_data",
                    "variant": "standard",
                }
            ],
        }
        mock_post.return_value = check_data
        user = get_user_model().objects.create_user(
            username='fred',
            first_name='Fred',
            last_name='Flintstone',
            email='fred@flintstone.com'
        )
        applicant = Applicant.objects.create_applicant(user, applicant_data)

        # 1. use the defaults.
        check = create_check(applicant, 'standard', ('identity', 'document'))
        mock_post.assert_called_once_with(
            'applicants/a9acefdf-3dc5-4973-aa78-20bd36825b50/checks',
            {'reports': [{'name': 'identity'}, {'name': 'document'}], 'type': 'standard'}
        )
        self.assertEqual(Check.objects.get(), check)
        # check we have two reports, and that the raw field matches the JSON
        # and that the parse method has run
        self.assertEqual(Report.objects.count(), 2)
        for r in check_data['reports']:
            # this will only work if the JSON has been parsed correctly
            report = Report.objects.get(onfido_id=r['id'])
            self.assertEqual(report.raw, r)

        # confirm that asserts guard input
        self.assertRaises(AssertionError, create_check, applicant, 'express', ('identity'))
        self.assertRaises(AssertionError, create_check, applicant, 'standard', 'identity')
        self.assertRaises(AssertionError, create_check, applicant, 'standard', None)

        # confirm that kwargs are merged in correctly
        check.delete()
        mock_post.reset_mock()
        check = create_check(applicant, 'standard', ('identity',), foo='bar')
        mock_post.assert_called_once_with(
            'applicants/a9acefdf-3dc5-4973-aa78-20bd36825b50/checks',
            {'reports': [{'name': 'identity'}], 'type': 'standard', 'foo': 'bar'}
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('onfido', '0006_rename_event_fields'),
    ]

    operations = [
        migrations.AlterField(
            model_name='event',
            name='onfido_id',
            field=models.CharField(help_text='The Onfido ID of the related resource.', max_length=40, verbose_name='Onfido ID'),
        ),
        migrations.AlterField(
            model_name='applicant',
            name='onfido_id',
            field=models.CharField(help_text='The id returned from the Onfido API.', unique=True, max_length=40, verbose_name='Onfido ID'),
        ),
        migrations.AlterField(
            model_name='check',
            name='onfido_id',
            field=models.CharField(help_text='The id returned from the Onfido API.', unique=True, max_length=40, verbose_name='Onfido ID'),
        ),
        migrations.AlterField(
            model_name='report',
            name='onfido_id',
            field=models.CharField(help_text='The id returned from the Onfido API.', unique=True, max_length=40, verbose_name='Onfido ID'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('onfido', '0007_rename_onfido_id_verbose_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='event',
            name='created_at',
            field=models.DateTimeField(help_text='The timestamp returned from the Onfido API.', null=True, blank=True),
        ),
        migrations.RenameField(
            model_name='event',
            old_name='created_at',
            new_name='completed_at',
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.10.4 on 2017-01-12 18:11
from __future__ import unicode_literals

from django.db import migrations, models


def set_event_received_at(apps, schema_editor):
    Event = apps.get_model("onfido", "Event")
    for event in Event.objects.all():
        event.received_at = event.completed_at
        event.save()


class Migration(migrations.Migration):

    dependencies = [
        ('onfido', '0011_add_indexes_to_status_fields'),
    ]

    operations = [
        migrations.AddField(
            model_name='event',
            name='received_at',
            field=models.DateTimeField(
                blank=True,
                null=True,
                help_text='The timestamp when the server received the event.',
            ),
        ),
        migrations.RunPython(set_event_received_at),
        migrations.AlterField(
            model_name='event',
            name='received_at',
            field=models.DateTimeField(
                help_text='The timestamp when the server received the event.'
            )
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import datetime
import logging

from dateutil.parser import parse as date_parse

from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _
from django.utils.timezone import now as tz_now

from .api import get
from .db.fields import JSONField
from .settings import scrub_report_data
from .signals import on_status_change, on_completion

logger = logging.getLogger(__name__)


class BaseModel(models.Model):

    """Base model used to set timestamps."""

    onfido_id = models.CharField(
        'Onfido ID',
        unique=True,
        max_length=40,
        help_text=_("The id returned from the Onfido API."),
    )
    created_at = models.DateTimeField(
        help_text=_("The timestamp returned from the Onfido API."),
        blank=True, null=True
    )
    raw = JSONField(
        help_text=_("The raw JSON returned from the API."),
        blank=True, null=True
    )

    class Meta:
        abstract = True

    @property
    def href(self):
        """Return the href from the raw JSON field."""
        return self.raw['href']

    def save(self, *args, **kwargs):
        """Save object and return self (for chaining methods)."""
        self.full_clean()
        super(BaseModel, self).save(*args, **kwargs)
        return self

    def parse(self, raw_json):
        """Parse the raw value out into other properties."""
        self.raw = raw_json
        self.onfido_id = self.raw['id']
        self.created_at = date_parse(self.raw['created_at'])
        return self

    def fetch(self):
        """
        Fetch the object JSON from the remote API.

        Named after the git operation - this will call the API for the
        latest JSON representation, and update the local fields, but will
        not save the updates. This is useful for inspecting the API response
        without making permanent changes to the object. It can also be used
        to interact with the API without saving an objects:

        >>> obj = Check(onfido_id='123').fetch()

        Returns the updated object (unsaved).

        """
        return self.parse(get(self.href))

    def pull(self):
        """
        Update the object from the remote API.

        Named after the git operation - this will call fetch(), and
        then save the object.

        Returns the updated object (saved).

        """
        return self.fetch().save()


class BaseQuerySet(models.QuerySet):

    """Custom queryset for models subclassing BaseModel."""

    def fetch(self):
        """Call fetch method on all objects in the queryset."""
        for obj in self:
            try:
                obj.fetch()
            except Exception:
                logger.exception("Failed to fetch Onfido object: %r", obj)

    def pull(self):
        """Call pull method on all objects in the queryset."""
        for obj in self:
            try:
                obj.pull()
            except Exception:
                logger.exception("Failed to pull Onfido object: %r", obj)


class BaseStatusModel(BaseModel):

    """Base class for models with a status and result field."""

    CHECK_STATUS_CHOICES = (
        ('in_progress', 'In progress'),
        ('awaiting_applicant', 'Awaiting applicant'),
        ('complete', 'Complete'),
        ('withdrawn', 'Withdrawn'),
        ('paused', 'Paused'),
        ('reopened', 'Reopened'),
    )
    REPORT_STATUS_CHOICES = (
        ('awaiting_data', 'Awaiting data'),
        ('awaiting_approval', 'Awaiting approval'),
        ('complete', 'Complete'),
        ('withdrawn', 'Withdrawn'),
        ('paused', 'Paused'),
        ('cancelled', 'Cancelled'),
    )
    STATUS_CHOICES = (
        ('Check', CHECK_STATUS_CHOICES),
        ('Report', REPORT_STATUS_CHOICES)
    )
    CHECK_RESULT_CHOICES = (
        ('clear', 'Clear'),
        ('consider', 'Consider')
    )
    REPORT_RESULT_CHOICES = (
        ('clear', 'Clear'),
        ('consider', 'Consider'),
        ('unidentified', 'Unidentified'),
    )
    RESULT_CHOICES = (
        ('Check', CHECK_RESULT_CHOICES),
        ('Report', REPORT_RESULT_CHOICES),
    )

    status = models.CharField(
        max_length=20,
        help_text=_("The current state of the check / report (from API)."),
        choices=STATUS_CHOICES,
        db_index=True,
        blank=True, null=True
    )
    result = models.CharField(
        max_length=20,
        choices=RESULT_CHOICES,
        help_text=_("The final result of the check / reports (from API)."),
        blank=True, null=True
    )
    updated_at = models.DateTimeField(
        blank=True, null=True,
        help_text=_("The timestamp of the most recent status change (from API).")
    )
    is_clear = models.NullBooleanField(
        default=None,
        help_text=_("True if the check / report is 'clear' (via API or manual override).")
    )

    class Meta:
        abstract = True

    def events(self):
        """Return queryset of Events related to this object."""
        return Event.objects.filter(onfido_id=self.onfido_id, resource_type=self._meta.model_name)

    def _override_event(self, user):
        """
        Create a fake override event for the object.

        This method uses the current object to create a fake Event payload,
        which is then parsed into an Event object, used to fake result
        transitions.

        NB the Event object isn't saved in this method.

        Args:
            user: User object, the person who is making the fake transition (
                used for auditing purposes).

        Returns a new (unsaved) Event that can be used to audit a fake transition.

        """
        payload = {
            "payload": {
                "action": "manual.override",
                "object": {
                    "completed_at": tz_now().isoformat(),
                    "href": self.href,
                    "id": self.onfido_id,
                    "status": self.status
                },
                "resource_type": self._meta.model_name,
                "user_id": user.id
            }

        }
        return Event().parse(payload)

    def update_status(self, event):
        """
        Update the status field of the object and fire signal(s).

        When the app receives an event callback from Onfido, we update
        the status of the relevant Check/Report, and then fire the
        signals that allow external apps to hook in to this event.

        If the update is a change to 'complete', then we fire a second
        signal - 'complete' is the terminal state change, and therefore
        of most interest to clients - typically the on_status_update
        signal would be registered for logging a complete history of
        changes, whereas the on_completion signal would be used to do
        something more useful - updating the status of the user, sending
        them an email etc.

        Args:
            event: Event object containing the update information

        Returns the updated object.

        """
        # we're doing a lot of marshalling from JSON to python, so this assert
        # just ensures we do actually have a datetime at this point
        assert isinstance(event.completed_at, datetime.datetime)
        # swap statuses around so we record old / new
        self.status, old_status = event.status, self.status
        self.updated_at = event.completed_at
        try:
            self.pull()
        except Exception:
            # even if we can't get latest, we should save the changes we
            # have already made to the object
            logger.warn("Unable to pull latest from Onfido: '%r'", self)
            self.save()
        on_status_change.send(
            self.__class__,
            instance=self,
            event=event.action,
            status_before=old_status,
            status_after=event.status
        )
        if event.status == 'complete':
            on_completion.send(self.__class__, instance=self)
        return self

    def parse(self, raw_json):
        """Parse the raw value out into other properties."""
        super(BaseStatusModel, self).parse(raw_json)
        self.result = self.raw['result']
        self.status = self.raw['status']
        if self.result == 'clear':
            self.is_clear = True
        return self

    def mark_as_clear(self, user):
        """
        Override the result field manually.

        When a Check / Report comes back as 'consider', and we need to get to
        some kind of answer ourselves (outside of Onfido). If we decided (offline)
        that the applicant is OK (e.g. their passport photo may be poor quality,
        so Onfido return 'consider', but we then take a view that it's correct)
        we need to mark them as cleared, so that they have the approval state.

        The important thing here is that we correctly audit the change - we don't
        just overwrite the result field, but we log the change, so that it can
        be seen. In order to do this we piggy-back on the event update process,
        and use a fake event (one that we create, not Onfido) to update the
        result.

        Args:
            user: User object, the person who is doing the overriding - this
                will typically be an admin user, doing it from the admin site.

        Returns the object itself, updated and saved.

        """
        self._override_event(user).save()
        self.is_clear = True
        self.save()
        return self


class ApplicantQuerySet(BaseQuerySet):

    """Custom Applicant queryset."""

    def create_applicant(self, user, raw):
        """Create a new applicant in Onfido from a user."""
        logger.debug("Creating new Onfido applicant from JSON: %s", raw)
        return Applicant(user=user).parse(raw).save()


class Applicant(BaseModel):

    """An Onfido applicant record."""

    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        help_text=_("Django user that maps to this applicant."),
        related_name='onfido_applicant'
    )

    objects = ApplicantQuerySet.as_manager()

    def __str__(self):
        return str(self.user)

    def __repr__(self):
        return "<Applicant id={} user_id={}>".format(
            self.id, self.user.id
        )


class CheckQuerySet(BaseQuerySet):

    """Check model manager."""

    def create_check(self, applicant, raw):
        """Create a new Check object from the raw JSON."""
        logger.debug("Creating new Onfido check from JSON: %s", raw)
        return Check(user=applicant.user, applicant=applicant).parse(raw).save()


class Check(BaseStatusModel):

    """The state of an individual check made against an Applicant."""

    CHECK_TYPE_CHOICES = (
        ('express', 'Express check'),
        ('standard', 'Standard check')
    )

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        help_text=_("The Django user (denormalised from Applicant to make navigation easier)."),  # noqa
        related_name='onfido_checks'
    )
    applicant = models.ForeignKey(
        Applicant,
        help_text=_("The applicant for whom the check is being made."),
        related_name='checks'
    )
    check_type = models.CharField(
        max_length=10,
        choices=CHECK_TYPE_CHOICES,
        help_text=_("See https://documentation.onfido.com/#check-types")
    )

    objects = CheckQuerySet.as_manager()

    def __str__(self):
        return "{} for {}".format(
            self.get_check_type_display().capitalize(),
            self.user
        )

    def __repr__(self):
        return "<Check id={} type='{}' user_id={}>".format(
            self.id,
            self.check_type,
            self.user.id
        )

    def parse(self, raw_json):
        """Parse the raw value out into other properties."""
        super(Check, self).parse(raw_json)
        self.check_type = self.raw['type']
        return self


class ReportQuerySet(BaseQuerySet):

    """Report model queryset."""

    def create_report(self, check, raw):
        """Create a new Report from the raw JSON."""
        logger.debug("Creating new Onfido report from JSON: %s", raw)
        return (
            Report(user=check.user, onfido_check=check)
            .parse(raw)
            .save()
        )


class Report(BaseStatusModel):

    """Specific reports associated with a Check."""

    REPORT_TYPE_CHOICES = (
        ('identity', 'Identity report'),
        ('document', 'Document report'),
        ('street_level', 'Street level report'),
        ('facial_similarity', 'Facial similarity report'),
        ('credit', 'Credit report'),
        ('criminal_history', 'Criminal history'),
        ('right_to_work', 'Right to work'),
        ('ssn_trace', 'SSN trace'),
    )

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        help_text=_("The Django user (denormalised from Applicant to make navigation easier)."),  # noqa
        related_name='onfido_reports'
    )
    onfido_check = models.ForeignKey(
        Check,
        help_text=_("Check to which this report is attached."),
        related_name='reports'
    )
    report_type = models.CharField(
        max_length=20,
        choices=REPORT_TYPE_CHOICES,
        help_text=_("The name of the report - see https://documentation.onfido.com/#reports")
    )

    objects = ReportQuerySet.as_manager()

    def __str__(self):
        return "{} for {}".format(
            self.get_report_type_display().capitalize(),
            self.user
        )

    def __repr__(self):
        return "<Report id={} type='{}' user_id={}>".format(
            self.id,
            self.report_type,
            self.user.id
        )

    def parse(self, raw_json):
        """
        Parse the raw value out into other properties.

        Before parsing the data, this method will call the
        scrub_report_data function to remove sensitive data
        so that it is not saved into the local object.

        """
        scrub_report_data(raw_json)
        super(Report, self).parse(raw_json)
        self.report_type = self.raw['name']
        return self


class Event(models.Model):

    """Used to record callback events received from the API."""

    onfido_id = models.CharField(
        'Onfido ID',
        max_length=40,
        help_text=_("The Onfido ID of the related resource."),
    )
    resource_type = models.CharField(
        max_length=20,
        help_text=_("The resource_type returned from the API callback.")
    )
    action = models.CharField(
        max_length=20,
        help_text=_("The event name as returned from the API callback.")
    )
    status = models.CharField(
        max_length=20,
        help_text=_("The status of the object after the event.")
    )
    completed_at = models.DateTimeField(
        help_text=_("The timestamp returned from the Onfido API."),
        blank=True, null=True
    )
    received_at = models.DateTimeField(
        help_text=_("The timestamp when the server received the event."),
    )
    raw = JSONField(
        help_text=_("The raw JSON returned from the API."),
        blank=True, null=True
    )

    class Meta:
        ordering = ['completed_at']

    def __str__(self):
        return "{} event occurred on {}.{}".format(
            self.action, self.resource_type, self.onfido_id
        )

    def __repr__(self):
        return "<Event id={} action='{}' onfido_id='{}.{}'>".format(
            self.id, self.action, self.resource_type, self.onfido_id
        )

    def _resource_manager(self):
        """Return the appropriate model manager for the resource_type."""
        assert self.resource_type in ('check', 'report'), (
            "Unknown resource type: {}".format(self.resource_type)
        )
        if self.resource_type == 'check':
            return Check.objects
        elif self.resource_type == 'report':
            return Report.objects

    @property
    def resource(self):
        """Return the underlying Check or Report resource."""
        return self._resource_manager().get(onfido_id=self.onfido_id)

    @property
    def user(self):
        """Return the user to whom the resource refers."""
        return self.resource.user

    def save(self, *args, **kwargs):
        """Save object and return self (for chaining methods)."""
        self.full_clean()
        super(Event, self).save(*args, **kwargs)
        return self

    def parse(self, raw_json):
        """Parse the raw value out into other properties."""
        self.raw = raw_json
        payload = self.raw['payload']
        self.resource_type = payload['resource_type']
        self.action = payload['action']
        obj = payload['object']
        self.onfido_id = obj['id']
        self.status = obj['status']
        self.completed_at = date_parse(obj['completed_at'])
        return self



# -*- coding: utf-8 -*-
"""Python 2/3 compatibility imports."""
try:
    from unittest import mock
except ImportError:
    import mock  # noqa

try:
    from urllib import parse as urlparse
except ImportError:
    import urlparse  # noqa



from django.core.management.base import BaseCommand
from ...models import Check, Report, Applicant


class Command(BaseCommand):

    help = "Pull all Check / Report objects."

    def add_arguments(self, parser):
        parser.add_argument(
            'model',
            choices=['applicant', 'check', 'report'],
            help="Type of model to sync"
        )
        parser.add_argument(
            '--filter',
            nargs='+',
            help="Status field values to filter on"
        )
        parser.add_argument(
            '--exclude',
            nargs='+',
            help="Status field values to exclude"
        )

    def handle(self, *args, **options):

        if options['model'] == 'check':
            model = Check
        elif options['model'] == 'report':
            model = Report
        elif options['model'] == 'applicant':
            model = Applicant

        filters = options['filter']
        excludes = options['exclude']

        objs = model.objects.all()
        objs = objs.filter(status__in=filters) if filters else objs
        objs = objs.exclude(status__in=excludes) if excludes else objs
        objs.pull()



# -*- coding: utf-8 -*-
"""onfido settings."""
from os import getenv

from django.conf import settings

# the API HTTP root url
API_ROOT = "https://api.onfido.com/v2/"

# API key from evnironment by default
API_KEY = (
    getenv('ONFIDO_API_KEY', None) or
    getattr(settings, 'ONFIDO_API_KEY', None)
)

# Webhook token - see https://documentation.onfido.com/#webhooks
WEBHOOK_TOKEN = (
    getenv('ONFIDO_WEBHOOK_TOKEN', None) or
    getattr(settings, 'ONFIDO_WEBHOOK_TOKEN', None)
)

# Set to False to turn off event logging
LOG_EVENTS = getattr(settings, 'ONFIDO_LOG_EVENTS', True)

# Set to True to bypass request verification (NOT RECOMMENDED)
TEST_MODE = getattr(settings, 'ONFIDO_TEST_MODE', False)


def DEFAULT_REPORT_SCRUBBER(raw):
    """Default report scrubber, removes breakdown and properties."""
    try:
        del raw['breakdown']
        del raw['properties']
    except KeyError:
        pass
    return raw

# function used to scrub sensitive data from reports
scrub_report_data = (
    getattr(settings, 'ONFIDO_REPORT_SCRUBBER', None) or
    DEFAULT_REPORT_SCRUBBER
)
